# -*- coding: UTF-8 -*-
"""Common tools."""

import os
import re
from multiprocessing import Process, Queue
from itertools import islice
import logging


def ifd(d, v1, v2, v3):
    """
    Conditionally select a value depending on the given dimension number.

    Args:
        d (int): Dimension number (1, 2, or 3).
        v1: The value if d = 1.
        v2: The value if d = 2.
        v3: The value if d = 3.

    Returns:
        v1, v2 or v3.

    """
    if d == 1:
        return v1
    elif d == 2:
        return v2
    elif d == 3:
        return v3
    else:
        raise ValueError("Invalid dimension: %s." % d)


def human_order_key(text):
    """
    Key function to sort in human order.

    """
    # This is based in http://nedbatchelder.com/blog/200712/human_sorting.html
    return [int(c) if c.isdigit() else c for c in re.split('(\d+)', text)]


def ensure_dir_exists(path):
    """
    Ensure a directory exists, creating it if needed.

    Args:
        path (str): The path to the directory.

    Raises:
        OSError: An error occurred when creating the directory.

    """
    if path:  # Empty dir (cwd) always exists
        try:
            # Will fail either if exists or unable to create it
            os.makedirs(path)
        except OSError:
            if os.path.exists(path):
                # Directory did [probably] exist
                pass
            else:
                # There was an error on creation, so make sure we know about it
                raise OSError("Unable to create directory " + path)


def get_dir_size(dir_path):
    """Get the size of a directory in bytes."""
    total_size = 0
    for root, dirs, files in os.walk(dir_path):
        for f in files:
            total_size += os.path.getsize(os.path.join(root, f))
    return total_size


def ensure_executable(path, all_users=None):
    """
    Ensure a file is executable.

    Args:
        path (str): the path to the file.
        all_users (bool): whether it should be make executable for the user or for all users.

    """
    st = os.stat(path)
    os.chmod(path, st.st_mode | (0o111 if all_users else 0o100))


class Call:
    """Objectization of a call to be made. When an instance is called, such a call will be made."""

    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.fn(*self.args, **self.kwargs)


def _caller(q):
    """Execute calls from a Queue until its value is 'END'."""
    os.setpgrp()  # To avoid inheriting SIGINT
    while True:
        data = q.get()
        if data == "END":
            break
        else:
            data()  # Call the supplied argument (lambda o Call instance)


class MPCaller:
    """
    MultiProcessing Caller. Makes calls using multiple subprocesses.
    
    Attributes:
        processes (list of multiprocessing.Process): Processes managed by the instance.
        
    """

    def __init__(self, num_threads=2):
        self._queue = Queue()
        self.processes = []
        self.spawn_threads(num_threads)
        self._ends_in_queue = 0

    def __repr__(self):
        return "MPCaller<%d threads, %d tasks in _queue>" % (
            len(self.processes), self._queue.qsize() - self._ends_in_queue)

    def spawn_threads(self, num_threads):
        """Create the required number of processes and add them to the caller.
        
        This does not remove previously created processes.
        
        """
        for _ in range(num_threads):
            t = Process(target=_caller, args=(self._queue,))
            t.daemon = True
            t.start()
            self.processes.append(t)

    def add_call(self, call):
        """
        Add a call to the instance's stack.
        
        Args:
            call (Callable): A function whose call method will be invoked by the processes. Consider using lambda
                             functions or a :class:`Call` instance.

        """
        self._queue.put(call)

    def wait_calls(self, blocking=True, respawn=False):
        """
        Ask all processes to consume the queue and stop after that.
        
        Args:
            blocking (bool): Whether to block the call, waiting for processes termination.
            respawn (bool): If blocking is True, this indicates whether to respawn the threads after the calls finish.
                            If blocking is not True this is ignored (no automatic respawn if non-blocking).

        """
        num_threads = len(self.processes)
        for _ in range(num_threads):
            self._queue.put("END")
            self._ends_in_queue += 1

        if blocking:
            for t in self.processes:
                t.join()
            self.processes = []
            self._ends_in_queue = 0
            if respawn:
                self.spawn_threads(num_threads)

    def abort(self, interrupt=False):
        """
        Remove all queued calls and ask processes to stop.

        Args:
            interrupt: If True, terminate all processes.

        """
        for _ in range(len(self.processes) + 1):
            self._queue.put("END")
            self._ends_in_queue += 1
        while True:
            # Consume all calls in queue
            data = self._queue.get()
            if data == "END":
                # If we consumed an END (we probably will), put it back
                self._queue.put("END")
                break
                # Otherwise do nothing
        if interrupt:
            for t in self.processes:
                t.terminate()
            # If the killed process was trying to use the Queue it could have corrupted
            # Just in case, create a new one
            self._queue = Queue()
            self._ends_in_queue = 0


def head(path, lines=10):
    """
    Get the first lines of a file.

    Args:
        path (str): Path to the file to read.
        lines (int): Number of lines to read.

    Returns:
        :obj:`list` of :obj:`str`: The lines found.

    """
    with open(path) as f:
        return list(islice(f, lines))


def tail(path, lines=1, _step=4098):
    """
    Get the last lines of a file.
    
    Args:
        path (str): Path to the file to read.
        lines (int): Number of lines to read.
        _step (int): Size of the step used in the search.

    Returns:
        :obj:`list` of :obj:`str`: The lines found.

    """
    # Adapted from glenbot's answer to:
    # http://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-with-python-similar-to-tail
    f = open(path, "r")
    lines_found = []
    block_counter = -1
    while len(lines_found) < lines:
        try:
            f.seek(block_counter * _step, os.SEEK_END)
        except IOError:  # either file is too small, or too many lines requested
            # read all and exit loop
            f.seek(0)
            lines_found = f.readlines()
            break

        lines_found = f.readlines()

        if len(lines_found) > lines:
            break
        block_counter -= 1

    return lines_found[-lines:]


logging.basicConfig(level=logging.INFO)
logger = logging.Logger("duat")



# -*- coding: UTF-8 -*-
"""Model OSIRIS configuration files."""

import copy
import re
from itertools import product
from functools import reduce  # py3 needs
from operator import mul

import numpy as np

from duat.common import ifd, logger


def _prod(iterable):
    return reduce(mul, iterable, 1)


def _val_to_fortran(val):
    """
    Transform a value to Fortran code.

    Args:
        val: The value to translate.

    Returns: 
        str: The Fortran code.

    """
    # Convert numpy types
    if isinstance(val, np.generic):
        val = val.item()

    t = type(val)
    if t is bool:
        return ".true." if val else ".false."
    elif t is int:
        return str(val)
    elif t is float:
        return ("%.6g" % val).replace("e", "d")
    elif t is str:
        return '"' + val + '"'
    raise TypeError("Unknown type: " + t)


def _fortran_to_val(val):
    """
    Transform a Fortran value to a Python one.

    Args:
        val (str): The value to translate.

    Returns:
        A Python representation.

    """
    if val[0] == val[-1] == '"':  # A string
        return val[1:-1]

    if val == ".true.":
        return True
    if val == ".false.":
        return False

    # Assume a number
    try:
        return int(val)
    except ValueError:
        pass

    # Assume a float
    val = val.replace("d", "e").replace("D", "e")
    try:
        return float(val)
    except ValueError:
        pass

    raise TypeError("Unable to parse: " + val)


def _par_to_fortran(name, val):
    """
    Transform a parameter to fortran code.

    Args:
        name (str): The name of the parameter.
        val: The value of the parameter.

    Returns:
        str: A string which assigns the value to the parameter.

    """
    # Convert numpy types
    if isinstance(val, np.generic):
        val = val.item()
    if isinstance(val, np.ndarray):
        val = val.tolist()

    t = type(val)
    if t in [bool, int, float, str]:
        return name + " = " + _val_to_fortran(val) + ","
    elif t in [list, tuple]:  # 1D or 2D or 3D

        if type(val[0]) in [list, tuple]:  # 2D or 3D
            if type(val[0][0]) in [list, tuple]:  # 3D
                s = ""
                for i, row in enumerate(val):
                    for j, row_row in enumerate(row):
                        s += name + "(1:" + str(len(row_row)) + ", " + str(i + 1) + ", " + str(
                            j + 1) + ") = " + ", ".join(
                            list(map(_val_to_fortran, row_row))) + ",\n  "
                return s
            else:  # 2D
                s = ""
                for i, row in enumerate(val):
                    s += name + "(1:" + str(len(row)) + ", " + str(i + 1) + ") = " + ", ".join(
                        list(map(_val_to_fortran, row))) + ",\n  "
                return s
        else:  # 1D
            l = len(val)
            return name + "(1:" + str(l) + ") = " + ", ".join(list(map(_val_to_fortran, val))) + ","
    raise TypeError


class MetaSection:
    """Something behaving like a set of sections."""

    def __str__(self):
        return self.to_fortran()

    def to_fortran(self):
        """Return a str with the fortran code representing the instance"""
        raise NotImplementedError()

    @classmethod
    def get_structure(cls, offset=0):
        """
        Get a string representing the structure of the class.

        This methods returns a string with lines in the format "- index (type)", where index is the one used to access
        a subitem and type is the class of that item.

        Args:
            offset: Two spaces times this number will be used to indent the returned string. This is used to generate a
                    multi-level description.

        Returns:
            (str): A representation of the structure of the class.

        """
        return ""


class Section(MetaSection):
    """The class defining a configuration block."""

    def __init__(self, name, param=None):
        self.name = name
        if param:
            self.pars = param.copy()
        else:
            self.pars = {}

    def __getitem__(self, ind):
        return self.pars[ind]

    def __setitem__(self, key, value):
        if value is None:
            if key in self.pars:
                self.pars.pop(key)
            # Otherwise ignore, raising no exception
        else:
            self.pars[key] = value

    def __contains__(self, key):
        return key in self.pars

    def set_par(self, name, val):
        """Add or update the value of a parameter"""
        self.pars[name] = val

    def set_pars(self, **kwargs):
        """Add or update some parameters using keyword arguments"""
        for key in kwargs:
            self.pars[key] = kwargs[key]

    def remove_par(self, name):
        """Remove the given parameter from the Section"""
        self.pars.pop(name)

    def to_fortran(self):
        s = self.name + "\n{\n"
        for p in self.pars:
            s += "  " + _par_to_fortran(p, self.pars[p]) + "\n"
        s += "}\n"
        return s


# TODO: Add a class derived from Section with prefixed parameters.


class SectionList(MetaSection):
    """
    Class defining a list of sections in a numerical order.

    Here 'section' refers to any subclass of `MetaSection`.
    
    """

    default_type = Section

    @classmethod
    def get_structure(cls, offset=0):
        s = ""
        t = Section if cls.default_type is None else cls.default_type
        s += " " * (offset * 2) + "- 0 (%s)\n" % (t.__name__,)
        s += t.get_structure(offset + 1)
        s += " " * (offset * 2) + "- 1 ...\n"
        return s

    def __init__(self, label=None, lst=None, default_type=None):
        """
        
        Args:
            label (str): a label inserted as a comment when generating fortran code.
            lst (list of MetaSection): a preexisting list of sections conforming this one. 
            default_type (str or class): the default type for implicit creation of content. If a str, a Section with the
                                         provided name will be created. Otherwise, a instance of the provided class with
                                         no parameters will be created.
        """
        self.label = label if label else ""
        self.lst = lst if lst else []
        if default_type is not None:
            self.default_type = default_type
        else:
            self.default_type = SectionList.default_type

    def __getitem__(self, ind):
        if ind < len(self.lst):
            return self.lst[ind]
        if self.default_type == Section:
            raise ValueError("A subsection cannot be implicitly added to the list due to generic default type.")
        if ind > len(self.lst):
            logger.warning("Implicitly creating more than one section in a list.")
        for i in range(len(self.lst), ind + 1):
            if isinstance(self.default_type, str):
                self.append_section(Section(self.default_type))
            else:
                # Note default_type is not Section here, no reason to expect incorrect call arguments in general.
                # Code analyzers may warn though.
                self.append_section(self.default_type())
        return self.lst[ind]

    def __setitem__(self, ind, value):
        if value is None:
            # TODO: Probably removing an item from a list with the ...= None notation is a bad practice
            raise NotImplementedError(
                "Removal from a list of Sections is not allowed. If you really need this, consider .lst.pop(number).")
        if ind < len(self.lst):
            self.lst[ind] = value
        elif ind == len(self.lst):
            self.lst.append(value)
        else:
            if self.default_type is None:
                raise ValueError("A subsection cannot be implicitly added to the list due to unknown type.")
            logger.warning("Implicitly added subsection(s).")
            for i in range(len(self.lst), ind):
                if isinstance(self.default_type, str):
                    self.append_section(Section(self.default_type))
                else:
                    # Note default_type is not Section here, no reason to expect incorrect call arguments in general.
                    # Code analyzers may warn though.
                    self.append_section(self.default_type())
            self.lst.append(value)

    def __iter__(self):
        for x in self.lst:
            yield x

    def __len__(self):
        return len(self.lst)

    def append_section(self, section):
        self.lst.append(section)

    def remove_section(self, number=-1):
        """Remove the section in the position given the index. By default, the last one."""
        self.lst.pop(number)

    def to_fortran(self):
        s = ("!---" + self.label + "\n") if self.label else ""
        for sec in self.lst:
            s += sec.to_fortran() + "\n"
        return s


class SectionOrdered(MetaSection):
    """
    Class defining a set of sections that must be outputted in a particular order given by a keyword related to them.

    Here 'section' refers to any subclass of `MetaSection`.
    
    """

    order = None
    types = None

    @classmethod
    def get_structure(cls, offset=0):
        s = ""
        for x in cls.order:
            t = cls.types[x]
            s += " " * (offset * 2) + "- %s (%s)\n" % (x, t.__name__)

            s += t.get_structure(offset + 1)

        return s

    def __init__(self, label=None, order=None, fixed=True, types=None):
        self.label = label if label else ""
        if not order:
            if fixed:
                logger.warning("A ConfigSectionOrdered instance with no order defined cannot be fixed.")
            self.fixed = False
            self.order = []
        else:
            self.fixed = fixed
            self.order = order

        self.types = types if types else {}
        self.subsections = {}

    def __getitem__(self, name):
        if isinstance(name, int):
            name = self.order[name]

        if name in self.subsections:
            return self.subsections[name]
        elif name in self.types:
            self.subsections[name] = self.types[name](name)
            return self.subsections[name]
        else:
            raise ValueError("Subsection %s cannot be implicitly created due to unknown type." % name)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            key = self.order[key]
        if value is None:
            if key in self.subsections:
                self.subsections.pop(key)
            # Otherwise ignore, raising no exception
        else:
            self.set_section(key, value)

    def __iter__(self):
        for x in self.order:
            yield x

    def __contains__(self, key):
        return key in self.subsections

    def set_section(self, name, section=None):
        """ Add or replace a section."""
        if section is None:  # No section was given
            section = Section(name) if "_list" not in name else SectionList(label=name)
        if isinstance(section, str):  # A string was given (abuse of notation)
            section = Section(section) if "_list" not in name else SectionList(label=section)
        if (name in self.subsections) or (name in self.order):
            self.subsections[name] = section
        else:
            if self.fixed:
                raise ValueError("Section %s not expected here. Valid names are %s" % (name, self.order))
            else:
                self.subsections[name] = section
                self.order.append(name)

    def remove_section(self, name):
        """Remove the section with the given name"""
        self.subsections.pop(name)

    def to_fortran(self):
        s = ("!---" + self.label + "\n") if self.label else ""
        for name in self.order:
            if name in self.subsections:
                s += self.subsections[name].to_fortran() + "\n"
        return s


# *****************************************
# Classes describing the semantic structure
# *****************************************


class Species(SectionOrdered):
    """Set of sections defining a species."""

    # Keep the default order as a class variable, but copy to the instance to allow modification
    order = ["species", "profile", "spe_bound", "diag_species"]
    types = {"species": Section, "profile": Section, "spe_bound": Section, "diag_species": Section}

    def __init__(self, label=None, dim=None):
        if isinstance(label, int):
            label = "Configuration for species " + str(label)
        SectionOrdered.__init__(self, label=label, order=Species.order, fixed=True, types=Species.types)

        # TODO: Move initialization to ConfigFileClass
        if dim is not None:
            self.set_section("species",
                             Section("species",
                                     {"num_par_max": 2048, "rqm": -1.0, "num_par_x": [2] * dim, "vth": [0.0] * 3,
                                      "vfl": [0.0] * 3}))

            default_profile = ifd(dim, {"fx": [[1., 1., 1., 1., 1., 1.]],
                                        "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.]]},
                                  # 2d
                                  {"fx": [[1., 1., 1., 1., 1., 1.], [1., 1., 1., 1., 1., 1.]],
                                   "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                         [0., 0.9999, 1.000, 2.000, 2.001, 10000.]]},
                                  # 3d
                                  {"fx": [[1., 1., 1., 1., 1., 1.], [1., 1., 1., 1., 1., 1.],
                                          [1., 1., 1., 1., 1., 1.]],
                                   "x": [[0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                         [0., 0.9999, 1.000, 2.000, 2.001, 10000.],
                                         [0., 0.9999, 1.000, 2.000, 2.001, 10000.]]}
                                  )
            self.set_section("profile", Section("profile", default_profile))

            self.set_section("spe_bound",
                             Section("spe_bound",
                                     {"type": ifd(dim, [[0, 0]], [[0, 0], [0, 0]], [[0, 0], [0, 0], [0, 0]])}))

            self.set_section("diag_species", Section("diag_species"))


class Cathode(SectionOrdered):
    """Set of sections defining a cathode."""

    # Keep the default order as a class variable, but copy to the instance to allow modification
    order = ["cathode", "species", "spe_bound", "diag_species"]
    types = {"cathode": Section, "species": Section, "spe_bound": Section, "diag_species": Section}

    def __init__(self, label=None):
        if isinstance(label, int):
            label = "Configuration for cathode " + str(label)
        SectionOrdered.__init__(self, label=label, order=Cathode.order, fixed=True, types=Cathode.types)


class Neutral(SectionOrdered):
    """Set of sections defining a neutral."""

    # Keep the default order as a class variable, but copy to the instance to allow modification
    order = ["neutral", "profile", "diag_neutral", "species", "spe_bound", "diag_species"]
    types = {"neutral": Section, "profile": Section, "diag_neutral": Section, "species": Section, "spe_bound": Section,
             "diag_species": Section}

    def __init__(self, label=None):
        if isinstance(label, int):
            label = "Configuration for neutral " + str(label)
        SectionOrdered.__init__(self, label=label, order=Neutral.order, fixed=True, types=Neutral.types)


# Types of lists of sections

class SpeciesList(SectionList):
    """List of sections defining the species"""
    default_type = Species

    def __init__(self, label="species"):
        SectionList.__init__(self, label=label)


class CathodeList(SectionList):
    """List of sections defining the cathodes"""
    default_type = Cathode

    def __init__(self, label="cathodes"):
        SectionList.__init__(self, label=label)


class NeutralList(SectionList):
    """List of sections defining the neutrals"""
    default_type = Neutral

    def __init__(self, label="neutrals"):
        SectionList.__init__(self, label=label)


class NeutralMovIonsList(SectionList):
    """List of sections defining the neutral moving ions"""

    def __init__(self, label="neutral moving ions"):
        # TODO: Write me
        raise NotImplementedError("Neutral moving ions are not yet implemented")
        # SectionList.__init__(self, label=label, default_type="neutral_mov_ions")


class ZpulseList(SectionList):
    """List of sections defining the laser pulses"""

    def __init__(self, label="zpulses"):
        SectionList.__init__(self, label=label, default_type="zpulse")


# Node sections


class SmoothCurrent(Section):
    def __init__(self, name):
        # Overwrite name with smooth
        Section.__init__(self, "smooth")


# Global input file model


class ConfigFile(SectionOrdered):
    """
    Set of Sections defining an input file.
    """
    order = ["simulation", "node_conf", "grid", "time_step", "restart", "space", "time", "el_mag_fld", "emf_bound",
             "smooth", "diag_emf", "particles", "species_list", "cathode_list", "neutral_list", "neutral_mov_ions_list",
             "collisions", "zpulse_list", "current", "smooth_current"]

    # TODO: Add antenna

    types = {"simulation": Section, "node_conf": Section, "grid": Section, "time_step": Section, "restart": Section,
             "space": Section, "time": Section, "el_mag_fld": Section, "emf_bound": Section, "smooth": Section,
             "diag_emf": Section, "particles": Section, "species_list": SpeciesList, "cathode_list": SpeciesList,
             "neutral_list": NeutralList, "neutral_mov_ions_list": NeutralMovIonsList, "collisions": Section,
             "zpulse_list": ZpulseList, "current": Section, "smooth_current": SmoothCurrent}

    def __init__(self, d=1, template="default"):
        """
        Create a default d-dimensional config file.

        Args:
            d(int): The number of dimensions (1, 2 or 3), used if a template is provided.
            template(str): A model for the initial configuration. Available values are:
                * "default": A basic configuration based on the examples.
                * "none": Absolutely no configuration.
            
        """
        SectionOrdered.__init__(self, order=ConfigFile.order, fixed=True, types=ConfigFile.types)

        if d not in [1, 2, 3]:
            raise TypeError("Invalid dimension")

        if template is None:
            template = "none"
        if not isinstance(template, str):
            raise TypeError("The template argument must be a string")
        template = template.lower()

        if template == "default":
            self["node_conf"] = Section("node_conf", {"node_number": [1] * d, "if_periodic": [True] * d})

            self["grid"] = Section("grid",
                                   {"coordinates": "cartesian", "nx_p": ifd(d, [1024], [32, 32], [10, 10, 10])})

            self["time_step"] = Section("time_step", {"dt": ifd(d, 0.07, 0.07, 0.05), "ndump": 10})

            self["space"] = Section("space", {"xmin": [0.0] * d, "xmax": ifd(d, [102.4], [3.2, 3.2], [1.0, 1.0, 1.0]),
                                              "if_move": [False] * d})

            self["time"] = Section("time", {"tmin": 0.0, "tmax": 7.0})

            self["emf_bound"] = Section("emf_bound",
                                        {"type": ifd(d, [[0, 0]], [[0, 0], [0, 0]], [[0, 0], [0, 0], [0, 0]])})

            self["particles"] = Section("particles", {"num_species": 2})

            self["species_list"] = SpeciesList()

            for i in [1, 2]:
                self["species_list"].append_section(Species(i, dim=d))

        elif template == "none":
            pass
        else:
            raise ValueError("Bad template parameter: %s" % template)

    @classmethod
    def from_string(cls, string):
        """Create and return an instance from a fortran code representation"""
        sim = ConfigFile(template="none")
        # Get (block_name, block_content) items
        blocks = re.findall(r"([^\n]*?)\n?{(.*?)\}", string, re.MULTILINE and re.DOTALL)

        for m in blocks:
            section = m[0]
            if section == "particles":
                # TODO: Account for other than regular species
                pass
            else:
                if section == "species":
                    l = len(sim["species_list"])
                    sim["species_list"][l] = Species(label=l + 1)
                    prev = sim["species_list"][l]
                    block = prev["species"]
                elif section in ["profile", "spe_bound", "diag_species"]:
                    block = prev[section]
                elif section == "zpulse":
                    sim["zpulse_list"].append_section(Section(name="zpulse"))
                    block = sim["zpulse_list"][-1]
                else:
                    block = sim[section]
                sizes = {}
                to_parse = {}
                # Get (parameter_name, access_modifier, parameter_value) items
                variables = re.findall(r"^\s*(\w*)\s*\(?(.*?)\)?\s*=\s* (.*?),$", m[1], re.MULTILINE)
                for variable in variables:
                    parameter, modifier, value = variable
                    if modifier:
                        dim = modifier.count(",") + 1
                        if dim == 1:
                            block[parameter] = list(map(lambda x: _fortran_to_val(x.strip()), value.split(",")))
                        elif dim == 2:
                            i = int(modifier.split(",")[-1])
                            if parameter in sizes:
                                sizes[parameter] = max(sizes[parameter], i)
                            else:
                                sizes[parameter] = i
                                to_parse[parameter] = {}
                            to_parse[parameter][i] = list(map(lambda x: _fortran_to_val(x.strip()), value.split(",")))
                        elif dim == 3:
                            raise NotImplementedError("Importing from dim 3 data is not yet implemented")
                        else:
                            raise ValueError("Parameter of dimension %d > 3" % dim)
                    else:
                        block[parameter] = _fortran_to_val(value)

                for (key, size) in sizes.items():  # inefficient on Py2
                    block[key] = [to_parse[key][i] for i in range(1, size + 1)]
        return sim

    @classmethod
    def from_file(cls, file_path):
        """Create and return an instance from a fortran code representation in a file"""
        with open(file_path, "r") as f:
            text = f.read()
        return ConfigFile.from_string(text)

    def _update_particles(self):
        """Update the particles Section with the currently set data"""
        self["particles"]["num_species"] = len(self["species_list"]) if "species_list" in self.subsections else 0
        self["particles"]["num_cathode"] = len(self["cathode_list"]) if "cathode_list" in self.subsections else 0
        self["particles"]["num_neutral"] = len(self["neutral_list"]) if "neutral_list" in self.subsections else 0
        self["particles"]["num_neutral_mov_ions"] = len(
            self["neutral_mov_ions_list"]) if "neutral_mov_ions_list" in self.subsections else 0

    def to_fortran(self):
        self._update_particles()
        return SectionOrdered.to_fortran(self)

    def write(self, path):
        """
        Save the config file to the specified path.

        Args:
            path: the path of the output file.

        Raises:
            OSError: if the file could not be opened.

        """
        with open(path, "w") as f:
            f.write(self.to_fortran())

    def clone(self):
        """
        Get a deep copy of the instance

        Returns:
            ConfigFile: a deep copy of this instance

        """
        return copy.deepcopy(self)

    def get_d(self):
        """
        Get the dimension of the configuration file.

        Returns:
            int: The dimension according to the mandatory xmax parameter in the space section. 0 if it could not be
                 found.
            
        """
        try:
            x_max = self["space"]["xmax"]
            return len(x_max) if isinstance(x_max, list) else 1

        except (ValueError, KeyError):  # space or xmax don't exist
            return 0

    def get_nodes(self):
        """
        Get the number of nodes required by the instance.

        Returns:
            int: Number of nodes.

        """
        try:
            nodes = self["node_conf"]["node_number"]
            return _prod(nodes) if isinstance(nodes, list) else nodes

        except (ValueError, KeyError):  # node_conf or node_number don't exist
            return 1


class Variation:
    """
    Represents a variation of a set of parameters in config files.
    
    Each parameter varies in list of values. The configuration files produced by this class take into account all
    combinations of values, i.e., the parameter space is given by the cartesian product.
    
    """

    def __init__(self, *args, epilog=None):
        """
        Create a Variation with the given parameters and values.
        
        Args:
            *args (2-:obj:`tuple` of :obj:`list`): Each argument must be a 2-tuple whose first elements is a list of str
                               or int which identifies the parameter in its section and a list of the values the
                               parameter will take. The list can be None to perform no action while passing the
                               parameter to the epilog function (see below).
            epilog (callable): A function of two arguments that will be called with the simulation and the list of
                               parameters when a config is being generated. This can be used for advanced modification,
                               for example, to set two parameters to a related value (like two species temperature).
        """
        self.parameters = args
        self._par_names = ["[dummy]" if p[0] is None else p[0][-1] for p in self.parameters]
        self.len_list = [len(p[1]) for p in self.parameters]
        self.epilog = epilog
        if epilog is None and all(p[0] is None for p in self.parameters):
            logger.warning("Trivial Variation generated. Did you forget an epilog?")

    def __repr__(self):
        return "Variation<%s (%s)>" % (" x ".join(self._par_names), "x".join([str(x) for x in self.len_list]))

    def get_generator(self, config):
        """
        Get a generator that produces ConfigFile objects following the Variation.
        
        Args:
            config (ConfigFile): The configuration where the Variation will be applied.

        Returns:
            generator: A generator which provides the ConfigFile instances.

        """
        paths = [p[0] for p in self.parameters]
        values = [p[1] for p in self.parameters]

        def gen():
            for value in product(*values):  # Value is multidimensional in general
                c = copy.deepcopy(config)
                for i, path in enumerate(paths):  # For the i-th thing to change
                    if path:  # If path is not None (non dummy parameter)
                        place = c  # Start from the root
                        for level in path[:-1]:  # ... access the leaf...
                            place = place[level]
                        place.set_par(path[-1], value[i])  # ... changing its value
                if self.epilog:
                    self.epilog(c, value)
                yield c

        return gen()

    def get_parameter_list(self):
        """
        Get a list with the parameter values in the same order that
        :func:`~duat.osiris.config.Variation.get_generator`.
        
        This method might be useful to post-process the results if the parameter space is simple.

        Returns:
            :obj:`list` of `tuple`: A list with the values of the parameters in the cartesian product order.

        """
        values = [p[1] for p in self.parameters]
        return list(product(*values))



# -*- coding: UTF-8 -*-
"""Plot OSIRIS-generated data."""

from __future__ import print_function
from __future__ import division

from glob import glob
import os
import re
from math import floor

import h5py
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import ticker
import matplotlib.colors as colors
from matplotlib.animation import FuncAnimation, FFMpegWriter

from duat.common import ensure_dir_exists, human_order_key, MPCaller, Call, logger


def _is_latex(s):
    """

    Decide if a string is a LaTeX expression.

    Args:
        s (str): The string to test.

    Returns:
        bool: True if it seems a LaTeX expression, False otherwise.

    """
    if not s:  # Empty string is not LaTeX
        return False
    return re.match("^.(_.)?$", s) or "\\" in s  # If in the form x, x_i, or if there is a backlash (command)


def _improve_latex(s):
    """Improve an OSIRIS-generated latex string using common rules."""
    # Descriptive subindexes in roman
    s2 = s.replace(r"\omega_p", r"\omega_{\mathrm{p}}")
    s2 = s2.replace("m_e", r"m_{\mathrm{e}}")
    # "Arbitrary units" in roman
    s2 = s2.replace("a.u.", r"\mathrm{a.u.}")
    return s2


def _create_label(name, units, latex_label=False):
    """
    Prepare a label for a plot
    
    Args:
        name (str): Text describing the magnitude 
        units (str): Text describing the units 
        latex_label (bool): Whether to use LaTeX 

    Returns:
        str: The desired label

    """
    if latex_label:
        if units:
            units = "$" + _improve_latex(units) + "$"
        # Names might be text or LaTeX. Try to guess
        if _is_latex(name):
            name = "$" + _improve_latex(name) + "$"

    if units:
        return "%s (%s)" % (name, units)
    else:
        return name


def _autonorm(norm, z):
    """
    Automatic options for color plot normalization.

    Args:
        norm (str or other): Description of the norm.
        z (matrix of numbers): Data.

    Returns:
        (`matplotlib.colors.Normalize`): A suitable normalize option (or None)

    """
    if isinstance(norm, str):
        norm = norm.lower()
    if norm == "lin":
        norm = None
    if norm == "log":
        masked_z = np.ma.masked_where(z <= 0, z)
        z_max = masked_z.max()
        z_min = masked_z.min()
        z_min = max(z_min, z_max / 1E9)
        norm = colors.LogNorm(vmin=z_min, vmax=z_max)
    return norm


def _fix_colorbar(cbar):
    # Manual fix for matplotlib's 8358 issue
    if isinstance(cbar.norm, colors.LogNorm) or isinstance(cbar.norm, colors.SymLogNorm):
        cbar.ax.minorticks_off()


class Diagnostic:
    """
    A OSIRIS diagnostic.

    Attributes:
        data_path (str): Path to the directory were the data is stored.
        data_name (str): A friendly name for the data.
        units (str): The name of the unit the magnitude is measured in.
        dt (float): The time step between snapshots of a consecutive number.
        t_0 (float): The time of the first snapshot.
        time_units (str): The name of the unit of time.
        file_list (list of str): List of h5 files, one per time snapshot.
        snapshot_list (list of int): List of integers identifying the snapshots. Multiply by dt to get time.
        keys (list of str): Names of the datasets in the Diagnostic, given in human order.
        axes (list of dict): Info of each axis in the Diagnostic.
        datasets_as_axis(dict): Info of datasets if treated as axes. WARNING: Only used with energy bins.
        shape (tuple): A tuple with:

            * list: The number of grid dimensions.
            * int: The number of datasets (excluding axes definition).
            * int: The number of snapshots in time.
            
    Note:
        The axes list is provided in the order of the numpy convention for arrays. This is the opposite of order used
        to label the axes in the hdf5 files. For example in a 2d array the first axes will be the labeled as AXIS2, and
        the second will be AXIS1. Unless the user makes use of other external tools to read the data, he/she can safely
        ignore this note.
            
    """

    def __init__(self, data_path):
        """
        Create a Diagnostic instance.

        Args:
            data_path (str): Path of the directory containing the diagnostic data.
        
        Raises:
            ValueError: If there is no data in `data_path`.
            
        """
        self.data_path = data_path
        self.file_list = glob(os.path.join(data_path, "*.h5"))
        if not self.file_list:
            raise ValueError("No diagnostic data found in %s" % data_path)
        self.file_list.sort(key=human_order_key)
        self.snapshot_list = list(
            map(lambda x: int((os.path.split(x)[1]).split(".h5")[0].split("-")[-1]), self.file_list))

        # Get info from first time snapshot
        with h5py.File(self.file_list[0], "r") as f:
            self.t_0 = f.attrs["TIME"][0]
            self.time_units = f.attrs["TIME UNITS"][0].decode('UTF-8')
            self.keys = self._get_keys(f)
            self.axes = self._get_axes(f, self.keys[0])
            self.units = f[self.keys[0]].attrs["UNITS"][0].decode('UTF-8')
            if len(self.keys) > 1:
                # Take a general name from the global attribute
                self.data_name = f.attrs["NAME"][0].decode('UTF-8')
                # Construct an axes-like object representing the dataset
                self.datasets_as_axis = {}
                # Guess name
                name = f[self.keys[0]].attrs["TAG"][0].decode('UTF-8')  # E.g., "Energy <=    0.00000000"
                name = re.match("(.*?) <", name).group(1)  # TODO: Catch errors, check syntax
                self.datasets_as_axis["NAME"] = name
                self.datasets_as_axis["LONG_NAME"] = name
                # Guess values
                dataset_axes = []
                number_pattern = r'[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?'  # A pattern matching floats
                for d in self.keys:
                    tag = f[d].attrs["TAG"][0].decode('UTF-8')
                    # tag is, e.g., '   0.00000000      < Energy <=    5.00000007E-02'
                    values = list(map(float, re.findall(number_pattern, tag)))
                    dataset_axes.append(sum(values) / len(values))  # So simple inequalities are represented somehow
                    # TODO: First and last datasets are not bounded intervals. Their only finite point is used as the x coordinate, which must be understood with caution
                self.datasets_as_axis["LIST"] = dataset_axes
                self.datasets_as_axis["MIN"] = dataset_axes[0]
                self.datasets_as_axis["MAX"] = dataset_axes[-1]
                self.datasets_as_axis["UNITS"] = "?"
            else:
                # Take a specific name from the dataset
                self.data_name = f[self.keys[0]].attrs["LONG_NAME"][0].decode('UTF-8')
                # No axes-like object in this case
                self.datasets_as_axis = None

        if len(self.file_list) < 2:
            self.dt = 0
        else:
            with h5py.File(self.file_list[1], "r") as f:
                self.dt = f.attrs["TIME"][0] / self.snapshot_list[1] - self.t_0

        self.shape = ([len(x["LIST"]) for x in self.axes], len(self.keys), len(self.snapshot_list))

    def __repr__(self):
        return "Diagnostic<%s %s>" % (self.data_name, str(self.shape))

    @staticmethod
    def _get_keys(file):
        """Get the dataset keys from an opened file."""
        keys = list(file.keys())
        if "AXIS" not in keys:
            raise ValueError("AXIS group not found.")
        keys.remove("AXIS")
        keys.sort(key=human_order_key)
        return keys

    @staticmethod
    def _get_axes(file, dataset_key=None):
        """Get the axes info."""
        if dataset_key is None:
            dataset_key = Diagnostic._get_keys(file)[0]
        axes = []
        for i, axis in enumerate(reversed(list(file["AXIS"].keys()))):
            ax = file["AXIS"][axis]
            data = {}
            for d in ["LONG_NAME", "UNITS", "NAME", "TYPE"]:
                data[d] = ax.attrs[d][0].decode('UTF-8')
            data["MIN"], data["MAX"] = ax[:]
            # TODO: Non linear axis support
            data["LIST"] = np.linspace(data["MIN"], data["MAX"], num=file[dataset_key].shape[i])
            axes.append(data)

        return axes

    def _clean_dataset_key(self, dataset_key):
        """Return the given dataset key as `str`, using human order if `int`. Might raise error or warning."""
        if isinstance(dataset_key, int):
            dataset_key = self.keys[dataset_key]
        elif isinstance(dataset_key, str):
            if dataset_key not in self.keys:
                raise ValueError("Dataset %s does not exist in the file." % dataset_key)
        elif dataset_key is None:
            if len(self.keys) != 1:  # Warn if implicitly selecting one among others.
                logger.warning("No dataset selected when multiple are available. Plotting the first one.")
            dataset_key = self.keys[0]
        else:
            raise TypeError("Unknown dataset type: %s", type(dataset_key))
        return dataset_key

    def _scaled_slice_to_slice(self, scaled_slice):
        return scaled_slice._get_slice(self.t_0, self.dt * len(self.snapshot_list), len(self.snapshot_list))

    def get_generator(self, dataset_selector=None, axes_selector=None, time_selector=None):
        """
        Get a generator providing data from the file.

        Calling this method returns a generator which, when called, will provide data for increasing times (unless
        modified by time_selector parameter). The data might be reduced either by selecting a position in an axis (or
        a dataset) or by using a function along some axis (or datasets), e.g., a sum.

        This data is provided as numpy arrays where the first axis refers to dataset coordinate (if present) and next
        to (non-reduced) axis in the order they are found in the files.

        Args:
            dataset_selector (str, int or callable): Instructions to reduce datasets. An int selects a dataset in human
                order, a str selects it by name. A function taking a list and returning a scalar can be used to reduce
                the data, e.g., sum, mean...

            axes_selector (tuple): Instructions to reduce axes data. It must be
                a tuple of the same length of the number axes or None to perform no reduction.
                Each element can be of the following types:

                    * int: Select the item in the given position.
                    * None: No reduction is performed in this axis.
                    * callable (default): Reduce the data along this axes using the given function (e.g., mean, max, sum...).


            time_selector (slice or ScaledSlice): A slice or ScaledSlice instance selecting the points in time to take.
                A slice selects times from the list returned by :func:`~duat.osiris.plot.Diagnostic.get_time_list`.
                A ScaledSlice chooses a slice that best represents a choice in terms of time units.

        Returns:
            generator: A generator which provides the data.

        """
        multiple_datasets = False  # If a dataset list is going to be returned
        if dataset_selector is not None:
            if self.shape[1] == 1:
                logger.warning("Single dataset found. Ignoring the provided dataset_selector.")

                def f_dataset_selector(f):
                    return f[self.keys[0]][:]
            else:
                if isinstance(dataset_selector, int):
                    dataset_selector = self.keys[dataset_selector]

                if isinstance(dataset_selector, str):  # If it was int or str
                    def f_dataset_selector(f):
                        return f[dataset_selector][:]
                else:  # Assumed function
                    def f_dataset_selector(f):
                        return np.apply_along_axis(dataset_selector, 0, [f[key][:] for key in self.keys])

        else:
            if self.shape[1] > 1:
                multiple_datasets = True

                def f_dataset_selector(f):
                    return np.array([f[key][:] for key in self.keys])
            else:
                def f_dataset_selector(f):
                    return f[self.keys[0]][:]

        if axes_selector is not None:
            if len(axes_selector) != len(self.axes):
                raise ValueError(
                    "Invalid axes_selector parameter. Length must be %d. Check the axes of the Diagnostic instance." % len(
                        self.axes))

            def f_axes_selector(x):
                offset = 1 if multiple_datasets else 0  # If multiple dataset, do not count its axis for reduction
                for i, sel in enumerate(axes_selector):
                    if sel is not None:
                        if isinstance(sel, int):
                            x = np.take(x, sel, axis=i - offset)
                        else:  # Assumed function
                            x = np.apply_along_axis(sel, i - offset, x)
                        offset += 1
                return x
        else:
            def f_axes_selector(x):
                return x

        if time_selector is not None:
            if isinstance(time_selector, ScaledSlice):
                time_selector = self._scaled_slice_to_slice(time_selector)
            elif not isinstance(time_selector, slice):
                logger.warning("Invalid time_selector parameter ignored. Use a slice or a ScaledSlice instead.")
                time_selector = None

        def gen():
            for file_name in (self.file_list[time_selector] if time_selector else self.file_list):
                with h5py.File(file_name, "r") as f:
                    data = f_dataset_selector(f)
                # Make sure to exit the context manager before yielding
                # h5py might accuse you of murdering identifiers if you don't!
                yield f_axes_selector(data)

        return gen()

    def get_axes(self, dataset_selector=None, axes_selector=None):
        """
        Get a dictionary with the info of the axes obtained as result of a given reduction.
        
        Args:
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.

        Returns:
            list of dict: Ordered list of the axes left by the reduction.
            
        """
        axes = []
        if dataset_selector is not None:
            if self.shape[1] == 1:
                logger.warning("Single dataset found. Ignoring the provided dataset_selector.")
                # Anyhow, datasets are reduced, so skip

        elif self.shape[1] > 1:
            axes.append(self.datasets_as_axis)

        if axes_selector is not None:
            if len(axes_selector) != len(self.axes):
                raise ValueError(
                    "Invalid axes_selector parameter. Length must be %d. Check the axes of the Diagnostic instance." % len(
                        self.axes))
            for i, sel in enumerate(axes_selector):
                if sel is None:
                    axes.append(self.axes[i])
        else:
            for a in self.axes:
                axes.append(a)
        return axes

    def get_time_list(self, time_selector=None):
        """
        Get the list of times obtained as a result of a given slice.
        
        Args:
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method. 

        Returns:
            :obj:`list` of :obj:`float`: The times resulting as a consequence of the slice.

        """
        if time_selector:
            if isinstance(time_selector, slice):
                # This could be improved to avoid generating unneeded values
                return [self.t_0 + self.dt * i for i in self.snapshot_list][time_selector]
            elif isinstance(time_selector, ScaledSlice):
                return self.get_time_list(self._scaled_slice_to_slice(time_selector))
            else:
                raise TypeError("time_selector must be a slice or a ScaledSlice")
        else:
            return [self.t_0 + self.dt * i for i in self.snapshot_list]

    def time_1d_animation(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                          dpi=200, fps=1, scale_mode="expand",
                          latex_label=True, interval=200):
        """
        Generate a plot of 1d data animated in time.
        
        If an output path with a suitable extension is supplied, the method will export it. Available formats are mp4
        and gif. The returned objects allow for minimal customization and representation. For example in Jupyter you
        might use `IPython.display.HTML(animation.to_html5_video())`, where `animation` is the returned `FuncAnimation`
        instance.
        
        Note:
            Exporting a high resolution animated gif with many frames might eat your RAM.

        Args:
            output_path (str): The place where the plot is saved. If "" or None, the plot is shown in matplotlib.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            interval (float): Delay between frames in ms. If exporting to mp4, the fps is used instead to generate the
                              file, although the returned objects do use this value.
            dpi (int): The resolution of the frames in dots per inch (only if exporting).
            fps (int): The frames per seconds (only if exporting to mp4).
            scale_mode (str): How the scale is changed through time. Available methods are:

                * "expand": The y limits increase when needed, but they don't decrease.
                * "adjust_always": Always change the y limits to those of the data.
                * "max": Use the maximum range from the beginning.

            latex_label (bool): Whether for use LaTeX code for the plot.
            
        Returns:
            (`matplotlib.figure.Figure`, `matplotlib.axes.Axes`, `matplotlib.animation.FuncAnimation`):
            Objects representing the generated plot and its animation.
            
        Raises:
            FileNotFoundError: If tried to export to mp4 but ffmpeg is not found in the system.

        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 1:
            raise ValueError("Expected 1 axis plot, but %d were provided" % len(axes))
        axis = axes[0]

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()
        fig.set_tight_layout(True)

        x_name = axis["LONG_NAME"]
        x_units = axis["UNITS"]
        y_name = self.data_name
        y_units = self.units

        ax.set_xlabel(_create_label(x_name, x_units, latex_label))
        ax.set_ylabel(_create_label(y_name, y_units, latex_label))

        # Plot the points
        x_min, x_max = axis["MIN"], axis["MAX"]
        plot_data, = ax.plot(axis["LIST"], next(gen))
        ax.set_xlim(x_min, x_max)

        if scale_mode == "max":
            # Get a list (generator) with the mins and maxs in each time step
            min_max_list = map(lambda l: [min(l), max(l)],
                               self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                                  time_selector=time_selector))
            f = lambda mins, maxs: (min(mins), max(maxs))
            y_min, y_max = f(*zip(*min_max_list))
            ax.set_ylim(y_min, y_max)

        time_list = self.get_time_list(time_selector)

        # Prepare a function for the updates
        def update(i):
            """Update the plot, returning the artists which must be redrawn."""
            try:
                new_dataset = next(gen)
            except StopIteration:
                logger.warning("Tried to add a frame to the animation, but all data was used.")
                return
            label = 't = {0}'.format(time_list[i])
            plot_data.set_ydata(new_dataset[:])
            ax.set_title(label)
            if not scale_mode or scale_mode == "max":
                pass
            elif scale_mode == "expand":
                prev = ax.get_ylim()
                data_limit = [min(new_dataset), max(new_dataset)]
                ax.set_ylim(min(prev[0], data_limit[0]), max(prev[1], data_limit[1]))
            elif scale_mode == "adjust_always":
                ax.set_ylim(min(new_dataset), max(new_dataset))
            return plot_data, ax

        anim = FuncAnimation(fig, update, frames=range(1, len(time_list) - 2), interval=interval)

        if not output_path:  # "" or None
            pass
        else:
            filename = os.path.basename(output_path)
            if "." in filename:
                extension = output_path.split(".")[-1].lower()
            else:
                extension = None
            if extension == "gif":
                anim.save(output_path, dpi=dpi, writer='imagemagick')
            elif extension == "mp4":
                metadata = dict(title=os.path.split(self.data_path)[-1], artist='duat', comment=self.data_path)
                writer = FFMpegWriter(fps=fps, metadata=metadata)
                with writer.saving(fig, output_path, dpi):
                    # Iterate over frames
                    for i in range(1, len(time_list) - 1):
                        update(i)
                        writer.grab_frame()
                    # Keep showing the last frame for the fixed time
                    writer.grab_frame()
            else:
                logger.warning("Unknown extension in path %s. No output produced." % output_path)

        plt.close()

        return fig, ax, anim

    def time_1d_colormap(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                         dpi=200, latex_label=True, cmap=None, norm=None, show=True, rasterized=True,
                         contour_plot=False):
        """
        Generate a colormap in an axis and the time.
    
        This function plots a magnitude depending on ONE spatial coordinate (hence the name) and on time as a colormap
        in the cartesian product of such a magnitude and the time.
        
        Note:
            For simple manipulation like labels or title you can make use of the returned tuple or a
            `matplotlib.pyplot.style.context`. More advanced manipulation can be done extracting the data with the
            :func:`~duat.osiris.plot.Diagnostic.get_generator` method instead.
    
        Args:
            output_path (str): The place where the plot is saved. If "" or None, the figure is not saved.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            dpi (int): The resolution of the file in dots per inch.
            latex_label (bool): Whether for use LaTeX code for the plot.
            cmap (str or `matplotlib.colors.Colormap`): The Colormap to use in the plot.
            norm (str or `matplotlib.colors.Normalize`): How to scale the colormap. For advanced manipulation, use some
                           Normalize subclass, e.g., colors.SymLogNorm(0.01). Automatic scales can be selected with
                           the following strings:

                           * "lin": Linear scale from minimum to maximum.
                           * "log": Logarithmic scale from minimum to maximum up to vmax/vmin>1E9, otherwise increasing vmin.


            show (bool): Whether to show the plot. This is blocking if matplotlib is in non-interactive mode.
            rasterized (bool): Whether the map is rasterized. This does not apply to axes, title... Note non-rasterized
                               images with large amount of data exported to PDF might challenging to handle.
            contour_plot (bool): Whether contour lines are plot instead of the density map.
                           
        Returns:
            (`matplotlib.figure.Figure`, `matplotlib.axes.Axes`): Objects representing the generated plot.
    
        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 1:
            raise ValueError("Expected 1 axis plot, but %d were provided" % len(axes))
        if len(self.file_list) < 2:
            raise ValueError("Unable to plot a colormap with only one time snapshot")
        axis = axes[0]

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()

        x_name = axis["LONG_NAME"]
        x_units = axis["UNITS"]
        y_name = "t"
        y_units = self.time_units
        title_name = self.data_name
        title_units = self.units

        ax.set_xlabel(_create_label(x_name, x_units, latex_label))
        ax.set_ylabel(_create_label(y_name, y_units, latex_label))

        time_list = self.get_time_list(time_selector)

        # Gather the points
        x_min, x_max = axis["MIN"], axis["MAX"]
        z = np.asarray(list(gen))

        norm = _autonorm(norm, z)

        plot_function = ax.contourf if contour_plot else ax.pcolormesh
        if rasterized:
            # Rasterizing in contourf is a bit tricky
            # Cf. http://stackoverflow.com/questions/33250005/size-of-matplotlib-contourf-image-files
            plot = plot_function(axis["LIST"], time_list, z, norm=norm, cmap=cmap, zorder=-9)
            ax.set_rasterization_zorder(-1)
        else:
            plot = plot_function(axis["LIST"], time_list, z, norm=norm, cmap=cmap)

        ax.set_xlim(x_min, x_max)
        ax.set_ylim(time_list[0], time_list[-1])

        ax.set_title(_create_label(title_name, title_units, latex_label))

        _fix_colorbar(fig.colorbar(plot))

        if output_path:  # "" or None
            plt.savefig(output_path, dpi=dpi)

        if show:
            plt.show()
        else:
            plt.close()

        return fig, ax

    def axes_2d_colormap(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                         dpi=200, latex_label=True, cmap=None, norm=None, show=True, rasterized=True,
                         contour_plot=False):
        """
        Generate a colormap in two axes.
        
        A single time snapshot must be selected with the time_selector parameter. For an animated version in time see 
        the :func:`~duat.osiris.plot.Diagnostic.time_2d_animation` method.


        Note:
            For simple manipulation like labels or title you can make use of the returned tuple or a
            `matplotlib.pyplot.style.context`. More advanced manipulation can be done extracting the data with the
            :func:`~duat.osiris.plot.Diagnostic.get_generator` method instead.

        Args:
            output_path (str): The place where the plot is saved. If "" or None, the figure is not saved.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            dpi (int): The resolution of the file in dots per inch.
            latex_label (bool): Whether for use LaTeX code for the plot.
            cmap (str or `matplotlib.colors.Colormap`): The Colormap to use in the plot.
            norm (str or `matplotlib.colors.Normalize`): How to scale the colormap. For advanced manipulation, use some
                           Normalize subclass, e.g., colors.SymLogNorm(0.01). Automatic scales can be selected with
                           the following strings:

                           * "lin": Linear scale from minimum to maximum.
                           * "log": Logarithmic scale from minimum to maximum up to vmax/vmin>1E9, otherwise increasing vmin.


            show (bool): Whether to show the plot. This is blocking if matplotlib is in non-interactive mode.
            rasterized (bool): Whether the map is rasterized. This does not apply to axes, title... Note non-rasterized
                               images with large amount of data exported to PDF might challenging to handle.
            contour_plot (bool): Whether contour lines are plot instead of the density map.

        Returns:
            (`matplotlib.figure.Figure`, `matplotlib.axes.Axes`): Objects representing the generated plot.

        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 2:
            raise ValueError("Expected 2 axes plot, but %d were provided" % len(axes))

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()

        x_name = axes[0]["LONG_NAME"]
        x_units = axes[0]["UNITS"]
        y_name = axes[1]["LONG_NAME"]
        y_units = axes[1]["UNITS"]
        title_name = self.data_name
        title_units = self.units

        ax.set_xlabel(_create_label(x_name, x_units, latex_label))
        ax.set_ylabel(_create_label(y_name, y_units, latex_label))

        time_list = self.get_time_list(time_selector)

        if len(time_list) == 0:
            raise ValueError("No time snapshot selected")
        if len(time_list) != 1:
            raise ValueError("A single time snapshot must be selected for this plot")

        # Gather the points
        x_min, x_max = axes[0]["MIN"], axes[0]["MAX"]
        y_min, y_max = axes[1]["MIN"], axes[1]["MAX"]
        z = np.transpose(np.asarray(list(gen)[0]))

        norm = _autonorm(norm, z)

        plot_function = ax.contourf if contour_plot else ax.pcolormesh
        if rasterized:
            # Rasterizing in contourf is a bit tricky
            # Cf. http://stackoverflow.com/questions/33250005/size-of-matplotlib-contourf-image-files
            plot = plot_function(axes[0]["LIST"], axes[1]["LIST"], z, norm=norm, cmap=cmap, zorder=-9)
            ax.set_rasterization_zorder(-1)
        else:
            plot = plot_function(axes[0]["LIST"], axes[1]["LIST"], z, norm=norm, cmap=cmap)

        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)

        ax.set_title(_create_label(title_name, title_units, latex_label))

        _fix_colorbar(fig.colorbar(plot))

        if output_path:  # "" or None
            plt.savefig(output_path, dpi=dpi)

        if show:
            plt.show()
        else:
            plt.close()

        return fig, ax

    def time_2d_animation(self, output_path=None, dataset_selector=None, axes_selector=None, time_selector=None,
                          dpi=200, fps=1, cmap=None, norm=None, rasterized=True, z_min=None,
                          z_max=None, latex_label=True, interval=200):
        """
        Generate a plot of 2d data as a color map which animated in time.

        If an output path with a suitable extension is supplied, the method will export it. Available formats are mp4
        and gif. The returned objects allow for minimal customization and representation. For example in Jupyter you
        might use `IPython.display.HTML(animation.to_html5_video())`, where `animation` is the returned `FuncAnimation`
        instance.

        Note:
            Exporting a high resolution animated gif with many frames might eat your RAM.

        Args:
            output_path (str): The place where the plot is saved. If "" or None, the plot is shown in matplotlib.
            dataset_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            axes_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            time_selector: See :func:`~duat.osiris.plot.Diagnostic.get_generator` method.
            interval (float): Delay between frames in ms. If exporting to mp4, the fps is used instead to generate the
                              file, although the returned objects do use this value.
            dpi (int): The resolution of the frames in dots per inch (only if exporting).
            fps (int): The frames per seconds (only if exporting to mp4).
            latex_label (bool): Whether for use LaTeX code for the plot.
            cmap (str or `matplotlib.colors.Colormap`): The Colormap to use in the plot.
            norm (str or `matplotlib.colors.Normalize`): How to scale the colormap. For advanced manipulation, use some
                           Normalize subclass, e.g., colors.SymLogNorm(0.01). Automatic scales can be selected with
                           the following strings:

                           * "lin": Linear scale from minimum to maximum.
                           * "log": Logarithmic scale from minimum to maximum up to vmax/vmin>1E9, otherwise increasing vmin.


            rasterized (bool): Whether the map is rasterized. This does not apply to axes, title... Note non-rasterized
                               images with large amount of data exported to PDF might challenging to handle.
        Returns:
            (`matplotlib.figure.Figure`, `matplotlib.axes.Axes`, `matplotlib.animation.FuncAnimation`):
            Objects representing the generated plot and its animation.
            
        Raises:
            FileNotFoundError: If tried to export to mp4 but ffmpeg is not found in the system.

        """
        if output_path:
            ensure_dir_exists(os.path.dirname(output_path))
        axes = self.get_axes(dataset_selector=dataset_selector, axes_selector=axes_selector)
        if len(axes) != 2:
            raise ValueError("Expected 2 axes plot, but %d were provided" % len(axes))

        gen = self.get_generator(dataset_selector=dataset_selector, axes_selector=axes_selector,
                                 time_selector=time_selector)

        # Set plot labels
        fig, ax = plt.subplots()
        fig.set_tight_layout(True)

        x_name = axes[0]["LONG_NAME"]
        x_units = axes[0]["UNITS"]
        y_name = axes[1]["LONG_NAME"]
        y_units = axes[1]["UNITS"]
        title_name = self.data_name
        title_units = self.units

        ax.set_xlabel(_create_label(x_name, x_units, latex_label))
        ax.set_ylabel(_create_label(y_name, y_units, latex_label))

        # Gather the points
        x_min, x_max = axes[0]["MIN"], axes[0]["MAX"]
        y_min, y_max = axes[1]["MIN"], axes[1]["MAX"]
        z = np.transpose(np.asarray(next(gen)))

        time_list = self.get_time_list(time_selector)
        if len(time_list) < 2:
            raise ValueError("At least two time snapshots are needed to make an animation")

        norm = _autonorm(norm, z)

        plot_function = ax.pcolormesh
        if rasterized:
            # Rasterizing in contourf is a bit tricky
            # Cf. http://stackoverflow.com/questions/33250005/size-of-matplotlib-contourf-image-files
            plot = plot_function(axes[0]["LIST"], axes[1]["LIST"], z, norm=norm, cmap=cmap, zorder=-9)
            ax.set_rasterization_zorder(-1)
        else:
            plot = plot_function(axes[0]["LIST"], axes[1]["LIST"], z, norm=norm, cmap=cmap)

        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)

        ax.set_title(_create_label(title_name, title_units, latex_label))

        _fix_colorbar(fig.colorbar(plot))

        # Prepare a function for the updates
        def update(i):
            """Update the plot, returning the artists which must be redrawn."""
            try:
                new_dataset = np.transpose(np.asarray(next(gen)))
            except StopIteration:
                logger.warning("Tried to add a frame to the animation, but all data was used.")
                return
            label = 't = {0}'.format(time_list[i])
            # BEWARE: The set_array syntax is rather problematic. Depending on the shading used in pcolormesh, the
            #         following might not work.
            plot.set_array(new_dataset[:-1, :-1].ravel())
            # For more details, check lumbric's answer to
            # https://stackoverflow.com/questions/18797175/animation-with-pcolormesh-routine-in-matplotlib-how-do-i-initialize-the-data
            ax.set_title(label)
            return plot, ax

        anim = FuncAnimation(fig, update, frames=range(1, len(time_list) - 2), interval=interval)

        if not output_path:  # "" or None
            pass
        else:
            filename = os.path.basename(output_path)
            if "." in filename:
                extension = output_path.split(".")[-1].lower()
            else:
                extension = None
            if extension == "gif":
                anim.save(output_path, dpi=dpi, writer='imagemagick')
            elif extension == "mp4":
                metadata = dict(title=os.path.split(self.data_path)[-1], artist='duat', comment=self.data_path)
                writer = FFMpegWriter(fps=fps, metadata=metadata)
                with writer.saving(fig, output_path, dpi):
                    # Iterate over frames
                    for i in range(1, len(time_list) - 1):
                        update(i)
                        writer.grab_frame()
                    # Keep showing the last frame for the fixed time
                    writer.grab_frame()
            else:
                logger.warning("Unknown extension in path %s. No output produced." % output_path)

        plt.close()

        return fig, ax, anim


def get_diagnostic_list(run_dir="."):
    """
    Create a list with the diagnostic found in the given run directory.

    Args:
        run_dir (str): The run directory.

    Returns:
        :obj:`list` of :obj:`Diagnostic`: List of the diagnostic found.

    """
    diagnostic_list = []
    # This might be not compatible with Windows due to the meaning of \ in patterns, but who cares...
    ms_folder = os.path.join(run_dir, "MS", "")  # Notice the empty string forces the route ending in the separator...
    for root, dirs, files in os.walk(ms_folder):
        if not dirs and files:  # Terminal directory with files in it
            diagnostic_list.append(Diagnostic(root))
    return diagnostic_list


def _pos_in_bin(x, a, b, n):
    """Find the bin where x is found in a mesh from a to b with n points (including both).
    The index starts at 0 and if x is in the mesh the interval to the right will be returned.
    The returned value can be outside [0, n-1] if x is not in [a, b).
    Note that for x=b, the returned value is n"""
    return floor((x - a) / (b - a) * n)


class ScaledSlice:
    """
    A slice described in simulation units (instead of list position).
    
    This object can be used to describe a time_selector parameter.
        
    """

    def __init__(self, start, stop, step=None):
        """
        Create a ScaledSlice instance.
        
        Args:
            start(float): Where the slice should start. Actual start will be before if needed. 
            stop(float): Where the slice should stop. The point is in general excluded, as usual in Python.
            step (float): The desired step of the slice. Actual step will be the biggest multiple of the mesh step which
                          is less than this one.
            
        """
        self.start = start
        self.stop = stop
        self.step = step

    def __repr__(self):
        if self.step:
            return "ScaledSlice<(%g, %g, %g)>" % (self.start, self.stop, self.step)
        else:
            return "ScaledSlice<(%g, %g)>" % (self.start, self.stop)

    def _get_slice(self, mesh_min, mesh_max, n_points):
        """Return a slice best approximating the instance in the given partition"""
        a = _pos_in_bin(self.start, mesh_min, mesh_max, n_points)
        if a < 0:
            a = 0
        if a >= n_points:
            a = n_points - 1
        b = _pos_in_bin(self.stop, mesh_min, mesh_max, n_points)
        if b < 0:
            b = 0
        if b >= n_points:
            b = n_points - 1
        if self.step:
            c = floor(self.step / ((mesh_max - mesh_min) / n_points))
            if c < 1:
                c = 1
            return slice(a, b, c)
        else:
            return slice(a, b)



# -*- coding: UTF-8 -*-
"""Run configuration files with OSIRIS."""

import re
import subprocess
from glob import glob
from os import path, remove, walk, listdir, environ
from shutil import copyfile
from time import sleep, time
from xml.etree import ElementTree

import psutil

from duat.plot import get_diagnostic_list as _get_diagnostic_list
from duat.common import ensure_dir_exists, ensure_executable, ifd, head, tail, logger, get_dir_size, human_order_key, \
    MPCaller, \
    Call
from duat.config import ConfigFile

# Path to osiris executables - guessed later in the code
osiris_1d = ""
"""str: Path to the osiris-1D.e file"""
osiris_2d = ""
"""str: Path to the osiris-2D.e file"""
osiris_3d = ""
"""str: Path to the osiris-3D.e file"""


def set_osiris_path(folder, warn=True):
    global osiris_1d, osiris_2d, osiris_3d
    if not path.isdir(folder):
        if warn:
            logger.warning("%s is not an existing folder." % folder)
        return

    r = path.join(folder, "osiris-1D.e")
    if path.isfile(r):
        osiris_1d = r
    elif warn:
        logger.warning("osiris-1D not found in %s" % folder)
    r = path.join(folder, "osiris-2D.e")
    if path.isfile(r):
        osiris_2d = r
    elif warn:
        logger.warning("osiris-2D not found in %s" % folder)
    r = path.join(folder, "osiris-3D.e")
    if path.isfile(r):
        osiris_3d = r
    elif warn:
        logger.warning("osiris-3D not found in %s" % folder)


class _memoized(object):
    def __init__(self, calls_to_live=5, time_to_live=2):
        self.cache = {}
        self.calls_to_live = calls_to_live
        self.time_to_live = time_to_live
        self._call_count = 1

    def __call__(self, func):
        def _memoized_function(*args):
            self.func = func
            now = time()
            try:
                value, last_update = self.cache[args]
                if self._call_count >= self.calls_to_live or now - last_update > self.time_to_live:
                    self._call_count = 1
                    raise AttributeError

                self._call_count += 1
                return value

            except (KeyError, AttributeError):
                value = self.func(*args)
                self.cache[args] = (value, now)
                return value

            except TypeError:
                return self.func(*args)

        return _memoized_function


# Memoize the process list so consecutive calls to _find_running_exe are faster (e.g., working with a list of Run).
@_memoized(calls_to_live=50, time_to_live=2)
def _get_process_list():
    return list(psutil.process_iter())


def _find_running_exe(exe):
    """Return the list of the pid of the processes of the argument executable"""
    candidates = []
    exe = path.abspath(exe)
    for proc in _get_process_list():
        try:
            pinfo = proc.as_dict(attrs=['pid', 'exe'])
        except psutil.NoSuchProcess:
            pass
        else:
            if pinfo["exe"] and pinfo['exe'] == exe:
                candidates.append(pinfo['pid'])
    return candidates


_qstat_available = True


# Memoize the process list so consecutive calls to _get_grid_jobs are faster (e.g., working with a list of Run).
# Also, if qstat is not available, stop trying to call it.
@_memoized(calls_to_live=50, time_to_live=2)
def _general_qstat():
    global _qstat_available
    if not _qstat_available:
        return None
    try:
        return subprocess.check_output("qstat -xml 2> /dev/null", shell=True)
    except subprocess.CalledProcessError:
        _qstat_available = False
        return None  # qstat not available


def _get_job_tree_text(job_tree, property):
    """Get some parameter of an ElementTree representing a job """
    # Check to prevent AttributeError: 'NoneType' object has no attribute 'text'
    try:
        return job_tree.find(property).text
    except AttributeError:
        return ""


def _get_grid_jobs():
    """
    Get information of active jobs in qstat
    
    Returns:
        list of dict: A list of dictionaries, each with info of a running process. Available keys include:
            job_number (int): The number identifying the job.
            script (str): Path to script launching the job.
            submission_time (int): Unix time when the job was submitted.
            cwd (str): Path of the current working directory.
            
            
    """
    output = _general_qstat()
    if not output:
        return None
    tree = ElementTree.fromstring(output)
    jobs = []
    for job in tree.iter('job_list'):
        job_number = job[0].text
        output = subprocess.check_output("qstat -j %s -xml" % job[0].text, shell=True)
        job_tree = ElementTree.fromstring(output)[0][0]  # First index is djob_info, second is element
        time_str = _get_job_tree_text(job_tree, "JB_submission_time")
        try:
            start_time = int(job_tree.find("JB_ja_tasks")[0].find("JAT_start_time").text)
        except (TypeError, AttributeError):
            # TypeError if JB_ja_tasks not in the tree (which will happen if not started)
            # AttributeError if JAT_start_time not in the subtree
            start_time = 0
        jobs.append({
            "job_number": int(job_number),
            "script": _get_job_tree_text(job_tree, "JB_script_file"),
            "submission_time": int(time_str) if time_str else 0,
            "start_time": start_time,
            "cwd": _get_job_tree_text(job_tree, "JB_cwd"),
        })
    return jobs


class Run:
    """
    An osiris run.
    
    Attributes:
        run_dir (str): Directory where the run takes place.
        total_steps (int): Amount of time steps in the simulation.
        running_mode (str): Can be "local" if a local process was found, "grid" if a grid job was found, or ""
                            otherwise. A simulation launch in the grid but running in the local machine will be tagged
                            as "local" (since you can access its process).
        processes (list of psutil.Process): If running_mode is "local", representation of the processes running the
                                            simulation.
        job (dict): If running_mode is "grid", information about the job of the simulation.
        
    """

    def __init__(self, run_dir):
        """
        Create a Run instance.
        
        Args:
            run_dir (str): Path where the OSIRIS run takes place. An os-stdin file must exist there.
             
        Raises:
            FileNotFoundError: If no os-stdin is found.
            
        """
        self.run_dir = run_dir
        self._run_dir_name = path.basename(run_dir)

        with open(path.join(run_dir, "os-stdin"), "r") as f:
            text = f.read()
        r = re.match(r".*time_step(.*?){(.*?)dt(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE)
        if not r:
            raise ValueError("No dt found in os-stdin.")
        dt = float(r.group(4))
        r = re.match(r".*time(.*?){(.*?)tmin(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE)
        t_min = float(r.group(4)) if r else 0.0
        r = re.match(r".*time(.*?){(.*?)tmax(.*?)=(.*?),(.*?)}", text, re.DOTALL + re.MULTILINE)
        if not r:
            raise ValueError("No tmax found in os-stdin. Default value 0.0 is trivial.")
        t_max = float(r.group(4))

        self.total_steps = int((t_max - t_min) // dt) + 1
        self.running_mode = ""
        self.processes = None
        self.job = None

        self._update()

    def get_config(self):
        """Return a ConfigFile instance parsing the os-stdin file"""
        return ConfigFile.from_file(path.join(self.run_dir, "os-stdin"))

    def get_status(self):
        """
        Return the status of the run.

        Returns:
            str: The status of the run. Possible values are:

                * "RUNNING": Either a process is running the simulation (running_mode="local") or a qstat job was found (running_mode="grid").
                * "FAILED": An error was detected in the OSIRIS output.
                * "FINISHED": The simulation has successfully finished its execution.
                * "INCOMPLETE": The simulation appears to have started but not running now. You should check the output to try to understand what happened. If restart information was sent, you might want to run "continue.sh" to resume the simulation.
                * "NOT STARTED": The files were created, but the simulation is not running. Note you will see this state if the Run is queued in a MPCaller instance.

        """
        if self._is_running():
            return "RUNNING"
        elif self._has_error():
            # The run started but failed
            return "FAILED"
        elif self._is_finished():
            # The run was finished
            return "FINISHED"
        elif self.current_step() >= 0:
            # The run started at some point but was not completed
            return "INCOMPLETE"
        else:
            # The run did not start
            return "NOT STARTED"

    def __repr__(self):
        status = self.get_status()
        if status == "RUNNING":
            return "Run<%s [RUNNING (%s/%d)]>" % (self._run_dir_name, self.current_step(), self.total_steps)
        elif status == "FAILED":
            # The run started but failed
            return "Run<%s [FAILED]>" % (self._run_dir_name,)
        elif status == "FINISHED":
            # The run was finished
            return "Run<%s> [FINISHED]" % (self._run_dir_name,)
        elif status == "INCOMPLETE":
            # The run started at some point but was not completed
            return "Run<%s> [INCOMPLETE (%s/%d)]" % (self._run_dir_name, self.current_step(), self.total_steps)
        elif status == "NOT STARTED":
            # The run did not start
            return "Run<%s> [NOT STARTED]" % (self._run_dir_name,)
        else:
            return "Run<%s> [???]" % (self._run_dir_name,)

    def _update(self):
        """Update the running info using what is found at the moment."""
        candidates = _find_running_exe(path.join(self.run_dir, "osiris"))

        try:
            if not candidates:  # No process running found
                self.processes = None
                # Try to find a job in queue
                jobs = _get_grid_jobs()
                if not jobs:  # Either no qstat or empty list
                    self.running_mode = ""
                else:
                    script_path = path.abspath(path.join(self.run_dir, "start.sh"))
                    valid_jobs = list(filter(lambda j: j["script"] == script_path, jobs))
                    if valid_jobs:
                        if len(valid_jobs) > 1:
                            logger.warning("More than one grid job was found for the run.")
                        self.job = valid_jobs[0]
                        self.running_mode = "grid"
                    else:  # No queued job
                        self.running_mode = ""

            else:
                self.processes = list(map(psutil.Process, candidates))
                self.running_mode = "local"

        except psutil.NoSuchProcess:
            # If the processes have died before processing was completed.
            self.processes = None
            self.running_mode = ""

    def current_step(self):
        """
        Find the current simulation step.
        
        Returns:
            int: The simulation step or -1 if it could not be found.

        """
        try:
            last_line = tail(path.join(self.run_dir, "out.txt"), 8)
        except FileNotFoundError:
            return -1
        if not last_line:  # Empty file
            return -1
        if re.search("now at  t", last_line[-1]):
            # Unless the line was incomplete, there should be a match with:
            a = re.match(r".* n = *(.*?)$", last_line[-1])
            if a:
                return int(a.group(1))
            # Otherwise, try the previous one
            a = re.match(r".* n = *(.*?)$", last_line[-2])
            if a:
                return int(a.group(1))
            else:
                return -1  # Some error exists in the file

        elif " Osiris run completed normally\n" in last_line:
            return self.total_steps
        else:
            return -1

    def _is_running(self):
        """Return True if the simulation is known to be running, or False otherwise."""
        # Public interface is given by get_status instead.
        self._update()
        return True if self.running_mode else False

    def _is_finished(self):
        # Public interface is given by get_status instead.
        if path.isfile(path.join(self.run_dir, "TIMINGS", "timings.001")):
            return True
        else:
            return False

    def _has_error(self):
        """Search for common error messages in the output file."""
        # Public interface is given by get_status instead.
        # TODO: Cache result if reached execution with no error
        try:
            # If there is something in the error file:
            if path.getsize(path.join(self.run_dir, "err.txt")) > 0:
                return True
        except FileNotFoundError:
            pass
        try:
            with open(path.join(self.run_dir, "out.txt"), "r") as f:
                text = f.read()
            # TODO: Depending on the file size, the following might be better. Investigate this.
            # text = "".join(head(path.join(self.run_dir, "out.txt"), 300)+tail(path.join(self.run_dir, "out.txt"), 10))

            # TODO: The commented option is slower (even if compiled) than this one. Investigate.
            if "(*error*)" in text or re.search("Error reading .* parameters", text) or re.search(
                    "MPI_ABORT was invoked", text):
                # if re.search("\(\*error\*\)|Error reading .* parameters|MPI_ABORT was invoked",text):
                return True
            else:
                return False
        except FileNotFoundError:
            return False

    def terminate(self):
        """
        Terminate the OSIRIS processes (if running).

        If runnning is "local", sends SIGINT to the processes. If "grid", calls qdel.

        Raises:
            subprocess.CalledProcessError: If using a grid and qdel fails.

        """
        self._update()
        if self.running_mode == "local":
            for process in self.processes:
                try:
                    process.terminate()
                except psutil.NoSuchProcess:
                    # The process has just terminated
                    # In multiprocess run this is likely to happen when other processes stops.
                    pass
        elif self.running_mode == "grid":
            subprocess.check_call("qdel %d" % self.job["job_number"], shell=True)
            pass
        else:
            logger.warning("Asked for termination of a Run not known to be running.")

    def kill(self):
        """
        Abruptly terminate the OSIRIS processes (if running).
        
        The :func:`~duat.osiris.run.Run.terminate` method should be used instead to perform a cleaner exit.

        If runnning is "local", sends SIGKILL to the processes. If "grid", calls qdel.

        Raises:
            subprocess.CalledProcessError: If using a grid and qdel fails.

        """
        self._update()
        if self.running_mode == "local":
            for process in self.processes:
                try:
                    process.kill()
                except psutil.NoSuchProcess:
                    # The process has just terminated
                    # In multiprocess run this is likely to happen when other processes stops.
                    pass
        elif self.running_mode == "grid":
            subprocess.check_call("qdel %d" % self.job["job_number"], shell=True)
            pass
        else:
            logger.warning("Asked for termination of a Run not known to be running.")

    def estimated_time(self):
        """
        Estimated time to end the simulation in seconds.
        
        The estimation uses a linear model and considers initialization negligible.
        For local runs, the start time of a process is used in the calculation.
        For grid runs, the start time of the job is used instead.

        If the run was resumed, the estimation will be wrong.
        
        Returns: 
            float: The estimation of the time to end the simulation or NaN if no estimation could be done.

        """
        self._update()
        if not self.running_mode:
            return 0 if self._is_finished() else float("nan")
        elif self.running_mode == "local":
            start = self.processes[0].create_time()
        elif self.running_mode == "grid":
            start = self.job["start_time"]
            if start == 0:
                # Queued, but not started
                return float("nan")
        else:
            logger.warning("Invalid running_mode attribute")
            return float("nan")
        current = self.current_step()
        if current <= 0:  # If not dumped yet or error
            return float('nan')
        else:
            elapsed = time() - start
            return elapsed * (self.total_steps / current - 1)

    def real_time(self):
        """Find the total time in seconds taken by the simulation if it has finished, otherwise returning nan."""
        try:
            # TODO: Update for resuming runs
            with open(path.join(self.run_dir, "TIMINGS", "timings.001"), "r") as f:
                text = f.read()
            r = re.match(r" Total time for loop was(?: *)(.*?)(?: *)seconds", text, re.DOTALL + re.MULTILINE)
            if not r:
                logger.warning("Bad format in timings file. The real time could not be read.")
                return float("nan")
            else:
                return float(r.group(1))
        except FileNotFoundError:
            return float("nan")

    def get_size(self):
        """Get the size of all run data in bytes."""
        return get_dir_size(self.run_dir)

    def get_diagnostic_list(self):
        """
        Create a list with the diagnostic found in the given Run.

        Returns:
            :obj:`list` of :obj:`Diagnostic`: List of the diagnostic found.

        """
        return _get_diagnostic_list(self.run_dir)


def open_run_list(base_path, filter=None):
    """
    Create a Run instance for each of the subdirectories in the given path.
    
    Args:
        base_path (str): Path where the runs are found.
        filter (str): Filter the directories using a UNIX-like pattern.

    Returns:
        list of `Run`: A list with the Run instances, ordered so their paths are in human order.

    """
    dir_list = listdir(base_path)
    if not dir_list:
        return []
    if filter is not None:
        filter_list = glob(path.join(base_path, filter))
        filter_list = [path.basename(x) for x in filter_list]
        dir_list = [x for x in dir_list if x in filter_list]
        if not dir_list:
            return []
    dir_list.sort(key=human_order_key)
    return [Run(x) for x in [path.join(base_path, y) for y in dir_list]]


def _execute_run(prefix, osiris_path, run_dir):
    """Execute and wait for a run to finish"""
    # Cf. run_config
    p = subprocess.Popen(prefix + osiris_path + " > out.txt 2> err.txt", shell=True, cwd=path.abspath(run_dir))
    p.wait()


def run_config(config, run_dir, prefix=None, clean_dir=True, blocking=None, force=None, mpcaller=None):
    """
    Initiate a OSIRIS run from a config instance.

    Args:
        config (`ConfigFile`): The instance describing the configuration file.
        run_dir (str): Folder where the run is carried.
        prefix (str): A prefix to run the command. If None, "mpirun -np X " will be used when a config with multiple
                      nodes is provided.
        clean_dir (bool): Whether to remove the files in the directory before execution.
        blocking (bool): Whether to wait for the run to finish.
        force (str): Set what to do if a running executable is found in the directory. Set to "ignore" to launch anyway,
                     possibly resulting in multiple instances running simultaneously; set to "kill" to terminate the
                     existing processes.
        mpcaller (MPCaller): An instance controlling multithreaded calls. If supplied, all calls will be handled by this
                     instance and the blocking parameter will be ignored.

    Returns:
        tuple: A Run instance describing the execution.

    """
    # Automatic mpirun# Find the needed amount of nodes
    n = config.get_nodes()
    if prefix is None:
        prefix = "mpirun -np %d " % n if n > 1 else ""
    elif prefix[-1] != " ":
        prefix += " "

    # Search for possibly running processes
    candidates = _find_running_exe(path.join(run_dir, "osiris"))
    if candidates:
        if force == "ignore":
            logger.warning("Ignored %d running exe found in %s" % (len(candidates), run_dir))
        elif force == "kill":
            logger.warning("Killing %d running exe found in %s" % (len(candidates), run_dir))
            for c in candidates:
                try:
                    psutil.Process(c).terminate()
                except psutil.NoSuchProcess:
                    pass  # If just ended
        else:
            logger.warning("Running exe found in %s. Aborting launch." % run_dir)
            return

    # Clean if needed
    if clean_dir:
        for root, dirs, files in walk(run_dir):
            for f in files:
                remove(path.join(root, f))

        for root, dirs, files in walk(run_dir):
            for f in files:
                logger.warning("Could not remove file %s" % f)

    # If the run is restartable, make the if_restart variable explicit
    if "restart" in config and "ndump_fac" in config["restart"] and config["restart"]["ndump_fac"]:
        if "if_restart" not in config["restart"]:
            config["restart"]["if_restart"] = False  # This is the default value

    # copy the input file
    ensure_dir_exists(run_dir)
    config.write(path.join(run_dir, "os-stdin"))

    # Copy the osiris executable
    osiris_path = path.abspath(path.join(run_dir, "osiris"))
    osiris = ifd(config.get_d(), osiris_1d, osiris_2d, osiris_3d)
    copyfile(osiris, osiris_path)
    ensure_executable(osiris_path)

    # Create a start.sh file to ease manual launch
    with open(path.join(run_dir, "start.sh"), 'w') as f:
        f.write("#!/bin/bash\n%s./osiris > out.txt 2> err.txt" % prefix)
    ensure_executable(path.join(run_dir, "start.sh"))

    # Create a continue.sh file to ease manual relaunch of aborted executions
    with open(path.join(run_dir, "continue.sh"), 'w') as f:
        f.write("#!/bin/bash"
                "\nsed -i -e \"s/if_restart = .false./if_restart = .true./g\" os-stdin"
                "\n./%s osiris >> out.txt 2>> err.txt" % prefix)
    ensure_executable(path.join(run_dir, "continue.sh"))

    if mpcaller is not None:
        run = Run(run_dir)
        # Set the run instance to update the process info when the call is made.
        mpcaller.add_call(Call(_execute_run, prefix, osiris_path, run_dir))
        return run
    else:
        proc = subprocess.Popen(prefix + osiris_path + " > out.txt 2> err.txt", shell=True, cwd=path.abspath(run_dir))
        if blocking:
            proc.wait()
        else:  # Sleep a little to check for quickly appearing errors and to allow the shell to start osiris
            sleep(0.2)

        # BEWARE: Perhaps under extreme circumstances, OSIRIS might have not started despite sleeping.
        # This could be solved calling the update method of the Run instance.
        # Consider this a feature instead of a bug :P

        run = Run(run_dir)

        # Try to detect errors checking the output
        if run._has_error():
            logger.warning(
                "Error detected while launching %s.\nCheck out.txt and err.txt for more information or re-run in console." % run_dir)
        return run


def run_config_grid(config, run_dir, prefix=None, run_name="osiris_run", remote_dir=None, clean_dir=True, prolog="",
                    epilog=""):
    """
    Queue a OSIRIS run in a compatible grid (e.g., Sun Grid Engine).

    Args:
        config (`ConfigFile`): The instance describing the configuration file.
        run_dir (str): Folder where the run will be carried.
        prefix (str): A prefix to run the command. If None, "mpirun -np X " will be used when a config with multiple
                      nodes is provided.
        run_name (str): Name of the job in the engine.
        remote_dir (str): If provided, a remote directory where the run will be carried, which might be only available
                          in the node selected by the engine. Note that if this option is used, the returned Run
                          instance will not access the remote_dir, but the run_dir. If the remote node is unable to
                          access the path (trying to create it if needed), OSIRIS will be started in the run dir and
                          errors will be logged by the queue system.
        clean_dir (bool): Whether to remove the files in the directory before execution.
        prolog (str): Shell code to run before calling OSIRIS (but once in the remote_dir if asked for).
        epilog (str): Shell code to run after calling OSIRIS.

    Returns:
        Run: A Run instance describing the execution.

    """
    # Clean if needed
    if clean_dir:
        for root, dirs, files in walk(run_dir):
            for f in files:
                remove(path.join(root, f))

        for root, dirs, files in walk(run_dir):
            for f in files:
                logger.warning("Could not remove file %s" % f)

    # Find the needed amount of nodes
    n = config.get_nodes()
    if prefix is None:
        prefix = "mpirun -np %d " % n if n > 1 else ""
    elif prefix[-1] != " ":
        prefix += " "

    # copy the input file
    ensure_dir_exists(run_dir)
    config.write(path.join(run_dir, "os-stdin"))

    # Copy the osiris executable
    osiris_path = path.abspath(path.join(run_dir, "osiris"))
    osiris = ifd(config.get_d(), osiris_1d, osiris_2d, osiris_3d)
    copyfile(osiris, osiris_path)
    ensure_executable(osiris_path)

    # Create a start.sh file with the launch script
    s = "".join(["#!/bin/bash\n#\n#$ -cwd\n#$ -S /bin/bash\n#$ -N %s\n" % run_name,
                 "#$ -pe smp %d\n" % n if n > 1 else "",
                 "#\n",
                 "NEW_DIR=%s\nmkdir -p $NEW_DIR\ncp -r . $NEW_DIR\ncd $NEW_DIR\n" % remote_dir if remote_dir else "",
                 prolog + "\n",
                 "\n%s./osiris > out.txt 2> err.txt\n" % prefix,
                 epilog + "\n"])

    with open(path.join(run_dir, "start.sh"), 'w') as f:
        f.write(s)
    ensure_executable(path.join(run_dir, "start.sh"))

    subprocess.Popen("qsub " + path.abspath(path.join(run_dir, "start.sh")), shell=True, cwd=path.abspath(run_dir))

    return Run(run_dir)


def run_variation(config, variation, run_base, caller=None, on_existing=None, **kwargs):
    """
    Make consecutive calls to :func:`~duat.osiris.run.run_config` with ConfigFiles generated from a variation.
    
    Args:
        config (`ConfigFile`): Base configuration file.
        variation (`Variation`): Description of the variations to apply.
        run_base (str): Path to the directory where the runs will take place, each in a folder named var_number.
        caller (int or `MPCaller`): If supplied, the calls will be managed by a MPCaller instance. If an int is provided
                                    an MPCaller with such a number of threads will be created. Provide an instance if
                                    interested in further controlling.
        on_existing (str): Action to do if a run of the variation exists. Only the names of the subfolders are used for
                           this purpose, which means the run could be different if the variation or the path have
                           changed. Set to "ignore" to leave untouched existing runs or set to "overwrite" to delete the
                           data and run a new instance. Default is like "ignore" but raising a warning.
        **kwargs: Keyword arguments to pass to :func:`~duat.osiris.run.run_config`

    Returns:
        list of Run: List with the Run instances in the variation directory.

    """
    r_list = []

    if on_existing is not None:
        if not isinstance(on_existing, str):
            raise ValueError("Invalid on_existing parameter")
        on_existing = on_existing.lower()
        if on_existing not in ["ignore", "overwrite"]:
            raise ValueError("Invalid on_existing parameter")

    if caller is None:
        for i, c in enumerate(variation.get_generator(config)):
            r = run_config(c, path.join(run_base, "var_" + str(i)), **kwargs)
            r_list.append(r)
    else:
        if isinstance(caller, int):
            _caller = MPCaller(caller)
        else:
            # Otherwise assume it was a MPCaller instance
            _caller = caller

        for i, c in enumerate(variation.get_generator(config)):
            var_dir = path.join(run_base, "var_" + str(i))
            if path.isfile(path.join(var_dir, "os-stdin")):
                # If the item existed
                if on_existing is None:
                    logger.warning("Skipping existing variation item " + var_dir)
                elif on_existing == "ignore":
                    pass
                else:  # overwrite
                    run_config(c, var_dir, mpcaller=_caller, **kwargs)
            else:
                # The item did not exist
                run_config(c, var_dir, mpcaller=_caller, **kwargs)
            r_list.append(Run(var_dir))

        if isinstance(caller, int):
            # If the MPCaller was created in this method, threads should die after execution
            _caller.wait_calls(blocking=False)
            # Nevertheless, processes seems not to be discarded until a new call to this method is made
    return r_list


def run_variation_grid(config, variation, run_base, run_name="os-var_", remote_dir=None, on_existing=None, **kwargs):
    """
    Make consecutive calls to :func:`~duat.osiris.run.run_config_grid` with ConfigFiles generated from a variation.

    Args:
        config (`ConfigFile`): Base configuration file.
        variation (`Variation`): Description of the variations to apply.
        run_base (str): Path to the directory where the runs will take place, each in a folder named var_number.
        run_name (str): Prefix to the name to use in the grid system.
        remote_dir (str): If provided, a remote directory where the runs will be carried, which might be only available
                          in the node selected by the engine. See :func:`~duat.osiris.run.run_config_grid`.
        on_existing (str): Action to do if a run of the variation exists. Only the names of the subfolders are used for
                           this purpose, which means the run could be different if the variation or the path have
                           changed. Set to "ignore" to leave untouched existing runs or set to "overwrite" to delete the
                           data and run a new instance. Default is like "ignore" but raising a warning.
        **kwargs: Keyword arguments to pass to :func:`~duat.osiris.run.run_config_grid`.

    Returns:
        list of Run: List with the Run instances in the variation directory.

    """
    r_list = []

    if on_existing is not None:
        if not isinstance(on_existing, str):
            raise ValueError("Invalid on_existing parameter")
        on_existing = on_existing.lower()
        if on_existing not in ["ignore", "overwrite"]:
            raise ValueError("Invalid on_existing parameter")

    for i, c in enumerate(variation.get_generator(config)):
        var_name = "var_" + str(i)
        var_dir = path.join(run_base, var_name)
        if path.isfile(path.join(var_dir, "os-stdin")):
            # If the item existed
            if on_existing is None:
                logger.warning("Skipping existing variation item " + var_dir)
            elif on_existing == "ignore":
                pass
            else:  # overwrite
                if remote_dir:
                    run_config_grid(c, var_dir, run_name=run_name + str(i),
                                    remote_dir=path.join(remote_dir, var_name), **kwargs)
                else:
                    run_config_grid(c, var_dir, run_name=run_name + str(i), **kwargs)
        else:
            # The item did not exist
            if remote_dir:
                run_config_grid(c, var_dir, run_name=run_name + str(i), remote_dir=path.join(remote_dir, var_name),
                                **kwargs)
            else:
                run_config_grid(c, var_dir, run_name=run_name + str(i), **kwargs)
        r_list.append(Run(var_dir))

    return r_list


# Try to guess the OSIRIS location:
_candidates = []
if "OSIRIS_PATH" in environ:
    _candidates.append(environ["OSIRIS_PATH"])
_candidates.append(path.join(path.expanduser("~"), "osiris", "bin"))
_candidates.append(path.join("usr", "local", "osiris", "bin"))

for t in _candidates:
    set_osiris_path(t, warn=False)
    if osiris_1d and osiris_2d and osiris_3d:
        break

if not (osiris_1d and osiris_2d and osiris_3d):
    if not (osiris_1d or osiris_2d or osiris_3d):
        logger.warning("Warning: no OSIRIS executables were found.")
    else:
        if not osiris_1d:
            logger.warning("Warning: osiris-1D.e not found.")
        if not osiris_2d:
            logger.warning("Warning: osiris-2D.e not found.")
        if not osiris_3d:
            logger.warning("Warning: osiris-3D.e not found.")

    logger.warning("Set the environment variable OSIRIS_PATH to a folder where the OSIRIS executables with names "
                   "osiris-1D.e and so on are found.\n"
                   "You can also use run.set_osiris_path or set the variables run.osiris_1d (and so on).")
